"use strict";

const fs = require('fs');
const dom = require('xmldom').DOMParser;
const util = require('util');
const querystring = require('querystring');
const crypto = require('crypto');
const http = require('http');
const https = require('https');
const path = require('path');
const ParserXML = require('./ParserXML.js');
const ParserColumn = require('./ParserColumn.js');
const EventEmitter = require('events');
const globule = require('globule');
const moment = require('moment');
const readline = require('readline');
const csvParse = require('csv-parse');
const iconv = require('iconv-lite');
const charset = require('jschardet');

const REFRESH_RATE = 200;

module.exports = class FileWatcher extends EventEmitter {

    /**
     * constructor
     *
     * @param {Object} config Configuration object
     * @param {String} key Key identifier
     */
    constructor(config, key) {

        super();

        this.config = config;
        this.key = key;
        this.scanTimeout = null;

        this.parserXML = new ParserXML(config, key);

        this.parserXML.on('error', function(e) {
            this.emit('error', {'message': e});
        }.bind(this));


        this.parserColumn = new ParserColumn(config, key);

        this.parserColumn.on('error', function(e) {
            this.emit('error', {'message': e});
        }.bind(this));

        this.watchingHashes = new Map();
        this.watching = new Map();

        this.retryFiles = {};

        this.checkFiles = new Map();
        if (!config.useHttp) {
            this.scanTimeout = setTimeout(this.scanPath.bind(this), REFRESH_RATE);
            globule.find(this.config.path).forEach((file) => {
                let stat = fs.statSync(file);
                this.watching.set(file, stat);
                
            });
            console.log('Watching:  ', key, '\t', 'File(s): '+this.config.path);
        } else {
            console.log('Listening: ', key, '\t', 'Uri: /'+this.config.url);
        }

        this.stopScan = false;

        var fullHost = this.config.host;
        this.port = fullHost.startsWith("http:") ? 80 : 443;
        this.host = fullHost.replace(/^https?\:\/\//, "");


        
    }

    kill() {
        this.stopScan = true;
    }

    handleError(e) {
        console.log(e);
    }

    /**
     * Scans the directory for new files;
     *
     */
    scanPath(){
        let newFiles = [], changedFiles = [];
        let checkFiles = new Map();
        try {
            globule.find(this.config.path).forEach(function(file) {
                let stat = fs.statSync(file);
                checkFiles.set(file, stat);
            });
        } catch (e) {
            this.emitError("[ERR] File Read Error: ", e);
        }

        for(let file of checkFiles.keys()){
            let file_stat = this.watching.get(file);
            // New file
            if (!file_stat){
                newFiles.push(file);
                continue;
            }

            // Changed file
            let stat = checkFiles.get(file);
            if (stat.mtime.getTime() != file_stat.mtime.getTime()){
                changedFiles.push(file);
                continue;
            }
        }


        setTimeout(() => {
            newFiles.forEach((file) => this.tryReadFile(file));
            changedFiles.forEach((file) => this.tryReadFile(file));

            this.watching = checkFiles;

            if (!this.stopScan)
                this.scanTimeout = setTimeout(this.scanPath.bind(this), REFRESH_RATE);
        }, 500);
    }


    tryReadFile (file) {

        var input;
        try {
            input = fs.readFileSync(file);
            input = iconv.decode(input, charset.detect(input).encoding);

        } catch (e) {
            //force utf-8 encoding and try again
            try {
                input = fs.readFileSync(file, 'utf-8');
            } catch (e) {

                if (typeof this.retryFiles[file] != 'undefined') {

                    if (this.retryFiles[file] > 3) {
                        //too much retries
                        this.retryFiles[file] = 0;
                        this.emitError("[ERR] Unable to read file, no permission? : " + file);
                        return;
                    }
                    this.retryFiles[file]++

                } else {
                    this.retryFiles[file] = 1;
                }
                setTimeout(this.tryReadFile.bind(this), 1000, file);
                return;
            }
        }
        return this.handleFileChange(input, file);

    }


    /**
     * handleFileChange
     *
     * @param {String} file filepath
     */
    handleFileChange(input, file) {

        let doc;
        let asc = [];

        //readfile

        if (!this.config.useHttp) {
            let fingerprint = crypto.createHash('md5').update(input).digest("hex");
            if (fingerprint == this.watchingHashes.get(file)) {
                return;
            }
            this.watchingHashes.set(file, fingerprint);
        }

        const filename = path.basename(file, '.xml');

        if (this.config.archive_file) {
            let extension = this.config.type == "xml" ? ".xml" : "";
            fs.writeFile(path.join((process.env.HOME || process.env.USERPROFILE), '/rmconnect/debug/' + path.basename(file, extension) + "_" + moment().format('YYYY-MM-DD-HHmmSSsss') + '.txt'), input, (err) => {
                if (err) console.error(err);
            });
        }

        console.log("START --------------------------------------");
        console.log("Parsing config: "+ this.config.key);
        console.log("Filename "+ filename);

        if (this.config.input_format == 'asc') {
            //ASC format for powergold

            var contents = fs.readFileSync(file, 'ASCII');
            var lines = contents.toString().split('\n');



            try {
                var blocks = this.parserColumn.parseMultiBlockFW(lines, filename);
            } catch (e) {
                this.emitError("[ERR] Item Parse Error: "+e.message);
                return;
            }

            for (var i in blocks) {
                this.makeAPIRequest(blocks[i]);
            }

            return;

        } else if (this.config.input_format == 'csv') {
            //CSV format for da future
            csvParse(input, this.config.csv_options, function(err, output) {
                if (typeof err != 'undefined' && err) {
                    this.emitError("[ERR] CSV Parse Error: "+err);
                    return;
                }

                try {
                    let blocks = [];
                    if (this.config.type == "now-on-air") {
                        this.emitError("[ERR] CSV Parse Error: Now On Air not supported");
                        return;
                        //return this.makeAPIRequest(this.parserColumn.parseNowOnAirCSV(output));
                    } else if (this.config.type == "playlist-hour") {
                        return this.makeAPIRequest(this.parserColumn.parsePlayListCSV(output, filename));
                    } else {
                        blocks = this.parserColumn.parseMultiBlockCSV(output, filename);
                        for (var i in blocks) {
                            this.makeAPIRequest(blocks[i]);
                        }
                        return;
                    }
                } catch (e) {
                    this.emitError("[ERR] Item Parse Error: "+e.message);
                    return;
                }
            }.bind(this));

            return;

        } else {//input format == xml

            //XML format for other playouts
            try {
                doc = new dom().parseFromString(input);
            } catch (e) {
                if (e.code == "EISDIR")
                    return;
                else
                    this.emitError("[ERR] XML Parse Error: "+e.message);
                return;
            }


            try {
                let sendData = false;
                let blocks = [];
                if (this.config.type == "now-on-air") {
                    sendData = this.parserXML.parseNowOnAirXML(doc, filename);
                } else if (this.config.type == "playlist-hour") {
                    sendData = this.parserXML.parsePlayListXML(doc, filename);
                } else {
                    if (typeof this.config.multi_block != "undefined") {
                        if (this.config.multi_block == "key") {
                            blocks = this.parserXML.parseMultiBlockXMLByKey(doc, filename);
                        } else if (this.config.multi_block == "parent") {
                            blocks = this.parserXML.parseMultiBlockXMLByParent(doc, filename);
                        }
                        for (var i in blocks) {
                            this.makeAPIRequest(blocks[i]);
                        }
                        return;
                    } else {
                        this.emitError("[ERR] Config Error: type is not now-on-air or playlist-day, but no key or parent in multi_block definition.")
                        return;
                    }
                }
                return this.makeAPIRequest(sendData);

             } catch (e) {
                 this.emitError("[ERR] Item Parse Error: "+e.message);
                 return;
             }

        }
    }


    /**
     * makeAPIRequest
     *
     * @param {Object} sendData data to send
     */
    makeAPIRequest(sendData, cb) {
        const buf = crypto.randomBytes(124);
        let postData = JSON.stringify(sendData);
        let nonce = buf.toString('hex');

        let ts = Math.round((new Date()).getTime() / 1000);
        let auth_string = util.format('%s:%s:%s:%s', ts, nonce, this.config.key, postData);
        let hmac = crypto.createHmac('sha256', this.config.secret);
        var options = {
            hostname: this.host,
            port: this.port,
            path: this.config.endpoint,
            method: 'POST',
            agent: require('../httpAgent.js'),
            headers: {
                'nonce': nonce,
                'timestamp': ts,
                'api-key': this.config.key,
                'hash': new Buffer(hmac.update(new Buffer(auth_string, 'utf-8')).digest('hex')).toString('base64'),
                'Content-Type': 'application/json',
                //'Content-Length': postData.length,
                'charset': 'utf8'
            }
        };
        console.log("OPTIONS", options);
        console.log("PAYLOAD", postData);

        var requester = http;

        if (this.port == 443) {
            requester = https;
        }

        let req = requester.request(options, (res) => {
            console.log('STATUS: ' + res.statusCode);
            console.log('HEADERS: ' + JSON.stringify(res.headers, null, 4));
            res.setEncoding('utf8');
            let body = "";
            res.on('data', (chunk) => {
                body += chunk;
                console.log('BODY: ' + chunk);
            });
            res.on('end', () => {
                if (cb) cb();
                console.log('END', '-'.repeat(40));
                this.emit('request', res.statusCode, body);
            });
        });

        req.on('error', function(e) {
            this.emitError('[ERR] problem with request: ' + e.message);
        }.bind(this));

        req.write(postData, 'utf8');
        req.end();

    }

    emitError(msg) {
        console.log(msg)
        this.emit('error', {'message' : msg});
    }

};
